from .classes.User import User as user
from .classes.Organization import Organization as organization
from .classes.Query import Query as query
from .classes.Table import Table as table
from .classes.File import File as file
from ._version import __version__

import pkg_resources  # part of setuptools
